/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btcore.c,v 1.9 2002/08/30 14:33:50 kds Exp $

   HCI Command Library

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#define __NO_VERSION__

#include <linux/config.h>
#include <linux/module.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/types.h>

/* Local Includes */
#include <affix/types.h>
#include <affix/btlist.h>
#include <affix/btdebug.h>
#include <affix/bluetooth.h>
#include <affix/btcore.h>


#define FILEBIT	DBHCILIB

static int	errno;

int __hci_open(char *name)
{
	int			fd, err;
	struct hci_open	cmd;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd == -1) {
		return -1;
	}

	if (name == NULL)
		cmd.cmd = HCI_OPEN_ANY | HCI_OPEN_KERNEL;
	else {
		cmd.cmd = HCI_OPEN_NAME | HCI_OPEN_KERNEL;
		strncpy(cmd.name, name, IFNAMSIZ);
	}

	err = btsys_ioctl(fd, BTIOC_OPEN_HCI, &cmd);
	if (err) {
		btsys_close(fd);
		return err;
	}

	return fd;
}

int __hci_open_id(int hd)
{
	int			fd, err;
	struct hci_open	cmd;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd == -1) {
		return -1;
	}

	cmd.cmd = HCI_OPEN_ID | HCI_OPEN_KERNEL;
	cmd.hd = hd;

	err = btsys_ioctl(fd, BTIOC_OPEN_HCI, &cmd);
	if (err) {
		btsys_close(fd);
		return err;
	}

	return fd;
}

int __hci_open_mgr(void)
{
	int			fd, err;
	struct hci_open	cmd;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd == -1) {
		return -1;
	}

	cmd.cmd = HCI_OPEN_MGR | HCI_OPEN_KERNEL;

	err = btsys_ioctl(fd, BTIOC_OPEN_HCI, &cmd);
	if (err) {
		btsys_close(fd);
		return err;
	}

	return fd;
}


#include <affix/hci_cmds.h>	//XXX

/* Exporting stuff */
EXPORT_SYMBOL(hci_open);
#if defined(CONFIG_AFFIX_PAN)
EXPORT_SYMBOL(HCI_SwitchRole);
EXPORT_SYMBOL(HCI_WriteScanEnable);
EXPORT_SYMBOL(HCI_ReadScanEnable);
EXPORT_SYMBOL(HCI_WriteClassOfDevice);
EXPORT_SYMBOL(HCI_WriteLinkSupervisionTimeout);
EXPORT_SYMBOL(HCI_ReadBDAddr);
EXPORT_SYMBOL(HCI_Inquiry);
EXPORT_SYMBOL(HCI_ChangeConnectionPacketType);
#endif

