/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btuart.c,v 1.2 2002/09/03 09:15:09 kds Exp $

   BTUART - physical protocol layer for Nokia Connectivity Card

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/


/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

#include <linux/config.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/ioctl.h>
#include <linux/file.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>

#include <linux/in.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/rtnetlink.h>

#include <asm/io.h>
#include <linux/serial_reg.h>

/* Local Includes */
#define	FILEBIT	DBUART

#include <affix/types.h>
#include <affix/btdebug.h>
#include <affix/hci.h>

#include "btuart.h"

#define BTUART_DELAY 500
/*************************** LOCAL FUNCTIONS DECL SECTION *********************/

/* Line discipline callback functions */
int  btuartld_init_generic(btuart_device *btuart);
int  btuartld_init_ncc(btuart_device *btuart);

void  btuartld_uninit_generic(btuart_device *btuart);
void  btuartld_uninit_ncc(btuart_device *btuart);

int  btuartld_open(struct tty_struct *tty);
void btuartld_close(struct tty_struct *tty);
int  btuartld_ioctl(struct tty_struct *, void *, int, void *);

int  btuartld_receive_room(struct tty_struct *tty);
void btuartld_write_wakeup(struct tty_struct *tty);

void btuartld_receive_buf(struct tty_struct *, const unsigned char *, char *, int);
void btuartld_receive_buf_generic(struct tty_struct *, const unsigned char *, char *, int);
void btuartld_receive_buf_ncc(struct tty_struct *, const unsigned char *, char *, int);

void defer_receive_buf_generic(void *data);
void defer_receive_buf_ncc(void *data);

/* Network callback functions */
int btuart_xmit_generic(struct sk_buff *skb, struct net_device *dev);
int btuart_xmit_ncc(struct sk_buff *skb, struct net_device *dev);


/*************************** LOCAL VARIABLES SECTION **************************/

struct tty_ldisc	btuart_ldisc;

btuart_dev_desc_t	btuart_dev_desc[] = {
{
	proto:			HCI_UART_TLP,
	flags:			AFFIX_UART_RI | AFFIX_UART_CTL | AFFIX_UART_LOW,
	speed:			0,
	type:			LMP_NOKIA,
	manfid:			MANFID_NOKIA,
	prodid:			PRODID_DTL1,
	hard_hdr_len:		TLP_HDR_LEN,
	net_xmit:		btuart_xmit_ncc,
	ldisc_init:		btuartld_init_ncc,
	ldisc_uninit:		btuartld_uninit_ncc,
	ldisc_receive_buf:	btuartld_receive_buf_ncc,
	defer_receive_buf:	defer_receive_buf_ncc
},
{
	proto:			HCI_UART_TLP,
	flags:			AFFIX_UART_RI | AFFIX_UART_CTL | AFFIX_UART_LOW,
	speed:			0,
	type:			LMP_NOKIA,
	manfid:			MANFID_NOKIA,
	prodid:			PRODID_DTL1_PROTO,
	hard_hdr_len:		TLP_HDR_LEN,
	net_xmit:		btuart_xmit_ncc,
	ldisc_init:		btuartld_init_ncc,
	ldisc_uninit:		btuartld_uninit_ncc,
	ldisc_receive_buf: 	btuartld_receive_buf_ncc,
	defer_receive_buf: 	defer_receive_buf_ncc
},
{
	proto:			HCI_UART_TLP,
	flags:			AFFIX_UART_RI | AFFIX_UART_CTL | AFFIX_UART_LOW,
	speed:			0,
	type:			LMP_NOKIA,
	manfid: 		MANFID_SOCKET,
	prodid: 		PRODID_SOCKET_BTCF1,
	hard_hdr_len: 		TLP_HDR_LEN,
	net_xmit:		btuart_xmit_ncc,
	ldisc_init:		btuartld_init_ncc,
	ldisc_uninit:		btuartld_uninit_ncc,
	ldisc_receive_buf:	btuartld_receive_buf_ncc,
	defer_receive_buf:	defer_receive_buf_ncc
},
{
	/* last - default descriptor */
	proto:			HCI_UART_H4,
	flags:			AFFIX_UART_CTL | AFFIX_UART_LOW,
	speed:			B115200,
	type:			-1,
	manfid:			0,
	prodid:			0,
	hard_hdr_len:		1,
	net_xmit:		btuart_xmit_generic,
	ldisc_init:		btuartld_init_generic,
	ldisc_uninit:		btuartld_uninit_generic,
	ldisc_receive_buf:	btuartld_receive_buf_generic,
	defer_receive_buf:	defer_receive_buf_generic
} };

#define NOF_UART_DEVS (sizeof(btuart_dev_desc) / sizeof(btuart_dev_desc[0]))


btlist_head_t	btuarts;

/*************************** FUNCTION DEFINITION SECTION **********************/

btuart_device *__btuart_lookup_device(kdev_t kdev)
{
	btuart_device	*btuart;

	btl_for_each (btuart, btuarts)
		if (kdev_same(btuart->hw.kdev, kdev))
			break;
	return btuart;
}

btuart_device *btuart_lookup_device(kdev_t kdev)
{
	btuart_device	*btuart;
	unsigned long	flags;

	read_lock_irqsave(&btuarts.lock, flags);
	btuart = __btuart_lookup_device(kdev);
	read_unlock_irqrestore(&btuarts.lock, flags);

	return btuart;
}

btuart_device *btuart_create(void)
{
	btuart_device	*btuart;
	unsigned long	flags;
	
	btuart = kmalloc(sizeof(*btuart), GFP_KERNEL);
	if (btuart == NULL)
		return NULL;
	memset(btuart, 0, sizeof(*btuart));

	/* initialize some attribues */
	btuart->magic = BTUARTLD_MAGIC;

  	skb_queue_head_init(&btuart->rx_queue);
	spin_lock_init(&btuart->lock);

	write_lock_irqsave(&btuarts.lock, flags);
	__btl_add_tail(&btuarts, btuart);
	write_unlock_irqrestore(&btuarts.lock, flags);

	return btuart;
}

void btuart_destroy(btuart_device *btuart)
{
	unsigned long	flags;

	write_lock_irqsave(&btuarts.lock, flags);
	__btl_unlink(&btuarts, btuart);
	skb_queue_purge(&btuart->rx_queue);
	kfree(btuart);
	write_unlock_irqrestore(&btuarts.lock, flags);
}


/* ************************************************  */

void inline btuart_wait_sync(btuart_device *btuart)
{
	struct timeval		ctime;
	suseconds_t		diff;

	DBPRT(DBCTRL, "btuart_wait_sync\n");
	
	if (test_and_clear_bit(0, &btuart->sync)) {
		get_fast_time(&ctime);
		diff = sub_timeval(&ctime, &btuart->sync_stamp);
		
		if (diff < NCC_SYNC_TIME) {
			DBPRT(DBCTRL, "udelay: %ld\n", NCC_SYNC_TIME - diff);

			udelay(NCC_SYNC_TIME - diff);
		}
	}
}

/*  Check if RI has changed. If it has the card can accept the next
    TLP packet, so we have to reenable the network sending queue.
    This type of flow control is Nokia specific. 
*/
void btuart_isr(int irq_num, void *dev_inst, struct pt_regs *regs)
{
#define UART_ISRMSG_CNT		200
	btuart_device *btuart = dev_inst;
	unsigned int port = btuart->hw.port;
	unsigned char uart_msr = inb(port + UART_MSR);

#ifdef CONFIG_AFFIX_DEBUG
	if (++btuart->irqcnt % UART_ISRMSG_CNT == 1)
		DBPRT(DBCTRL, "cnt=%d  RI=%d\n", 
			btuart->irqcnt, 
			uart_msr & UART_MSR_RI ? 1 : 0);
#endif

	/*  Note that we cannot check if its our interrupt. The isr in serial.c
	    has already been executed so the interrupt pending bit may be 0.
	*/
	if (!hci_running(btuart->dev))
		return;
	if (btuart->uart_ri_latch ^ (uart_msr & UART_MSR_RI)) {
		DBPRT(DBCTRL, "Reenable sending queue\n");
		btuart->uart_ri_latch = uart_msr & UART_MSR_RI;
		hci_wake_queue(btuart->dev, btuart->flowmask);
	}
}

int btuartld_open(struct tty_struct *tty)
{
	btuart_device	*btuart;
	int		err = -EEXIST;

	DBFENTER;

	BTASSERT(tty != NULL, return -EEXIST;);
  
	MOD_INC_USE_COUNT;

	/* First make sure we're not already connected. */
	btuart = (btuart_device *) tty->disc_data;
	if (btuart && btuart->magic == BTUARTLD_MAGIC) {
		BTERRORF("Already opened\n");
		goto err1;
	}

	btuart = btuart_lookup_device(tty->device);
	if (btuart == NULL) {
		BTERRORF("Device not found\n");
		err = -ENODEV;
		goto err1;
	}

	if ( btuart->magic != BTUARTLD_MAGIC) {
		BTERRORF("WRONG MAGIC\n");
		goto err1;
	}
	
  	btuart->tty = tty;
	tty->disc_data = btuart;

	/* flush all internal buffers */
	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);
  
	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);
  
	DBFEXIT;
	return 0;

 err1:
	tty->disc_data = NULL;
	MOD_DEC_USE_COUNT;
	return err;
}

void btuartld_close(struct tty_struct *tty)
{
	btuart_device	*btuart = (btuart_device *) tty->disc_data;
	
	DBFENTER;

	/* First make sure we're connected. */
	BTASSERT(btuart && btuart->magic == BTUARTLD_MAGIC, return;);

	if (btuart == NULL) return;

	/* Stop tty */
	tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
	tty->disc_data = 0;
  
	btuart->tty = NULL;

	MOD_DEC_USE_COUNT;

	DBFEXIT;
}

/*
 * Function btuartld_ioctl (tty, file, cmd, arg)
 *
 *     The Swiss army knife of system calls :-)
 *
 */
int  btuartld_ioctl(struct tty_struct *tty, void *file, int cmd, void *arg)
{
#ifdef CONFIG_AFFIX_DEBUG
	btuart_device	*btuart = (btuart_device *) tty->disc_data;
#endif

	DBFENTER;

	BTASSERT(btuart != NULL, return -ENODEV;);
	BTASSERT(btuart->magic == BTUARTLD_MAGIC, return -EBADR;);

	return n_tty_ioctl(tty, (struct file *) file, cmd, 
				   (unsigned long) arg);

	DBFEXIT;
	return 0;
}

/*
 * Function btuartld_receive_room (tty)
 *
 *    Used by the TTY to find out how much data we can receive at a time
 * 
*/
int  btuartld_receive_room(struct tty_struct *tty)
{
	DBFENTER;

	return 65536;  /* We can handle an infinite amount of data. :-) */
}

/*
   if we change status from disable to enable.. -> mark_bh
*/
int btuart_receive_ncc_control(btuart_device *btuart, struct sk_buff *skb)
{
	__u8	flowmask = *(__u8*)skb->data;

	DBFENTER;

	DBPRT(DBRECV, "-->> CONTROL packet\n");

	/* transition to active state */
	if ((btuart->flowmask & 0x07) == 0 && (flowmask & 0x07) != 0)
		hci_wake_queue(btuart->dev, btuart->flowmask);

	btuart->flowmask = flowmask;

	if (flowmask & NCC_SYNC) {
		/* wait SYNC_TIME */
		set_bit(0, &btuart->sync);
		get_fast_time(&btuart->sync_stamp);
	}

	//netdev_state_change(btuart->dev);	/* just let HCI check state */

	kfree_skb(skb);
	DBFEXIT;
	return 0;
}

int btuart_receive_data_generic(btuart_device *btuart, struct sk_buff *skb)
{
	hci_skb_cb		*cb = (hci_skb_cb*)skb->cb;
	__u8			*hci_type = (__u8*)skb->data;
	struct net_device	*dev = btuart->dev;
	
	DBFENTER;

	DBPRT(DBRECV, "HCI packet, len: %d\n", skb->len);
	DBPRTPKT(DBRECV, skb->data, skb->len);

	/* remove uart header */
	skb_pull(skb, 1);

	/* now skb contain pure HCI packet */

	switch (*hci_type) {

	case UART_HCI_ACL:
		DBPRT(DBRECV, "ACL\n");
		cb->type = HCI_ACL;
		break;

	case UART_HCI_SCO:
		DBPRT(DBRECV, "SCO\n");
		cb->type = HCI_SCO;
		break;

	case UART_HCI_EVENT:
		DBPRT(DBRECV, "EVENT\n");
		cb->type = HCI_EVENT;
		break;

	default:
		DBPRT(DBRECV, "BAD UART HCI PACKET TYPE: %#02x\n", *hci_type);
		/* may be free skb */
		kfree_skb(skb);
		return -1;
	}

	skb->dev = dev;
	skb->mac.raw = (void*)hci_type;
	skb->protocol = __constant_htons(ETH_P_BLUETOOTH);
	skb->pkt_type = PACKET_HOST;
	
	hci_rx(skb);	/* send to upper protocol layer */
	
	DBFEXIT;
	return 0;
}


int btuart_receive_data_ncc(btuart_device *btuart, struct sk_buff *skb)
{
	hci_skb_cb	*cb = (hci_skb_cb*)skb->cb;
	tlp_hdr_t	*tlp = (tlp_hdr_t*)skb->data;
	struct net_device	*dev = btuart->dev;
	
	DBFENTER;

	DBPRT(DBRECV, "TLP packet received, length: %d\n", skb->len);
	DBPRTPKT(DBRECV, skb->data, skb->len);

	/* remove TLP header */
	skb_pull(skb, TLP_HDR_LEN);

	/* now skb contain pure HCI packet */

	switch (tlp->type) {

	case TLP_HCI_ACL:
		DBPRT(DBRECV, "-->> ACL packet\n");
		cb->type = HCI_ACL;
		break;

	case TLP_HCI_SCO:
		DBPRT(DBRECV, "-->> SCO packet\n");
		cb->type = HCI_SCO;
		break;

	case TLP_HCI_EVENT:
		DBPRT(DBRECV, "-->> EVENT packet\n");
		cb->type = HCI_EVENT;
		break;

	default:
		DBPRT(DBRECVE, "BAD TLP TYPE: %#02x\n", tlp->type);
		/* may be free skb */
		kfree_skb(skb);
		return -1;
	}

	skb->dev = dev;
	skb->mac.raw = (void*)tlp;
	skb->protocol = __constant_htons(ETH_P_BLUETOOTH);
	skb->pkt_type = PACKET_HOST;
	
	hci_rx(skb);	/* send to upper protocol layer */
	
	DBFEXIT;
	return 0;
}

void defer_receive_buf_generic(void *data)
{
	btuart_device		*btuart = data;
	struct sk_buff		*skb;

	DBFENTER;

	BTASSERT(btuart && btuart->magic == BTUARTLD_MAGIC, return;);

	while ((skb = skb_dequeue(&btuart->rx_queue)))
		btuart_receive_data_generic(btuart, skb);

	DBFEXIT;
}

void defer_receive_buf_ncc(void *data)
{
	btuart_device		*btuart = data;
	struct sk_buff		*skb;

	DBFENTER;

	BTASSERT(btuart && btuart->magic == BTUARTLD_MAGIC, return;);

	while ((skb = skb_dequeue(&btuart->rx_queue)))
		btuart_receive_data_ncc(btuart, skb);

	DBFEXIT;
}

/*
  should be used with low_latency == 1
*/
void btuartld_receive_buf(struct tty_struct *tty, const unsigned char *cp, char *fp, int count)
{
	btuart_device	*btuart = tty->disc_data;
	
	if (btuart->dev && !hci_device_present(btuart->dev))
		return;
	if (btuart->dev_desc.ldisc_receive_buf)
		btuart->dev_desc.ldisc_receive_buf(tty, cp, fp, count);
}

void btuartld_receive_buf_generic(struct tty_struct *tty, const unsigned char *cp, char *fp, int count)
{
	btuart_device		*btuart = tty->disc_data;
	const unsigned char	*ptr;
	int			rlen, lmin;
	int			i = 0;

	DBFENTER;

	DBPRT(DBRECV, "We've received %d bytes\n", count);
	DBPRTPKT(DBRECV, (__u8*)cp, count);

	if (btuart == NULL) {
		BTERROR(" No btuart device, btuart == NULL\n");
		return;
	}

	for (i = 0; i < count; i++) {
		if (fp[i])
			BTNOTICEF("FLAG ERROR: %#02x\n", fp[i]);
	}

	ptr = cp;
	rlen = count;

	while (rlen) {
		if (btuart->hlen == 0) {	/* expect new packet */
			btuart->hdr[0] = *ptr;
			btuart->rpos = 1;
			ptr += 1;
			rlen -= 1;
			
			switch (btuart->hdr[0]) {
			case UART_HCI_ACL:
				btuart->hlen = 1 + sizeof(HCI_ACL_Packet_Header);
				break;
			case UART_HCI_SCO:
				btuart->hlen = 1 + sizeof(HCI_SCO_Packet_Header);
				break;
			case UART_HCI_EVENT:
				btuart->hlen = 1 + sizeof(HCI_Event_Packet_Header);
				break;
			default:
				BTERROR("Invalid HCI packet type %#02x\n", btuart->hdr[0]);
				break;
			}

		} else if (btuart->rpos < btuart->hlen) {
			lmin = btmin(btuart->hlen - btuart->rpos, rlen);
			memcpy(btuart->hdr + btuart->rpos, ptr, lmin);
			btuart->rpos += lmin;
			ptr += lmin;
			rlen -= lmin;

			if (btuart->rpos >= btuart->hlen) {
				switch (btuart->hdr[0]) {
				case UART_HCI_ACL:
					btuart->plen = btuart->hlen +
						__btoh16(((HCI_ACL_Packet_Header*)&btuart->hdr[1])->Length);
					break;
				case UART_HCI_SCO:
					btuart->plen = btuart->hlen +
						((HCI_SCO_Packet_Header*)&btuart->hdr[1])->Length;
					break;
				case UART_HCI_EVENT:
					btuart->plen = btuart->hlen +
						((HCI_Event_Packet_Header*)&btuart->hdr[1])->Length;
					break;
				}

				btuart->rskb = dev_alloc_skb(btuart->plen);
				if (!btuart->rskb) {
					BTERRORF("No memory for skbuff\n");
					/* No free memory */
					btuart->hlen = 0;
					return;
				}
				/* copy tlp header */
				memcpy(skb_put(btuart->rskb, btuart->hlen), btuart->hdr, btuart->hlen);
			}

		} else {
			if (btuart->rskb->len < btuart->plen) {
				lmin = btmin(btuart->plen - btuart->rskb->len, rlen);
				memcpy(skb_put(btuart->rskb, lmin), ptr, lmin);
				ptr += lmin;
				rlen -= lmin;
				if (btuart->rskb->len >= btuart->plen) {
					if (tty->low_latency) {
						skb_queue_tail(&btuart->rx_queue, btuart->rskb);
						tasklet_hi_schedule(&btuart->rx_task);
					} else
						btuart_receive_data_generic(btuart, btuart->rskb);

					btuart->hlen = 0;
					btuart->rskb = NULL;
				}
			}
		}
	}
}

/*
  should be used with low_latency == 1
*/
void btuartld_receive_buf_ncc(struct tty_struct *tty, const unsigned char *cp, char *fp, int count)
{
	btuart_device		*btuart = tty->disc_data;
	const unsigned char	*ptr;
	int			rlen, lmin;
	int			i = 0;

	DBFENTER;

	//DBPRT(DBRECV, "We've received %d bytes\n", count);

	if (btuart == NULL) {
		BTERROR(" No btuart device, btuart == NULL\n");
		return;
	}

	for (i = 0; i < count; i++) {
		if (fp[i])
			BTNOTICEF("FLAG ERROR: %#02x\n", fp[i]);
	}

	ptr = cp;
	rlen = count;

	while (rlen) {
		if (btuart->pos <  TLP_HDR_LEN) {
			lmin = btmin(TLP_HDR_LEN - btuart->pos, rlen);
			memcpy(((__u8*)&btuart->tlp)+btuart->pos, ptr, lmin);
			btuart->pos += lmin;
			ptr += lmin;
			rlen -= lmin;
			
			if (btuart->pos >= TLP_HDR_LEN) {
				btuart->tlp.length = __btoh16(btuart->tlp.length);
				btuart->plen = TLP_HDR_LEN + btuart->tlp.length +
					TLP_PAD_LEN(btuart->tlp.length);
				btuart->rskb = dev_alloc_skb(btuart->plen);
				if (btuart->rskb == NULL) {
					BTERRORF("No memory for skbuff\n");
					/* No free memory */
					return;
				}
				/* copy tlp header */
				memcpy(skb_put(btuart->rskb, TLP_HDR_LEN), &btuart->tlp, TLP_HDR_LEN);
			}
		} else {
			if (btuart->rskb->len < btuart->plen) {
				lmin = btmin(btuart->plen - btuart->rskb->len, rlen);
				memcpy(skb_put(btuart->rskb, lmin), ptr, lmin);
				ptr += lmin;
				rlen -= lmin;
				if (btuart->rskb->len >= btuart->plen) {
					/* send skb to upper layer */
					/* remove PAD byte if it exists */
					if (TLP_PAD_LEN(btuart->tlp.length)) {
						btuart->rskb->tail--;
						btuart->rskb->len--;
					}
					if (btuart->tlp.type == TLP_CONTROL) {
						skb_pull(btuart->rskb, TLP_HDR_LEN);
						btuart_receive_ncc_control(btuart, btuart->rskb);
					} else {
						if (tty->low_latency) {
							skb_queue_tail(&btuart->rx_queue, btuart->rskb);
							tasklet_hi_schedule(&btuart->rx_task);
						} else
							btuart_receive_data_ncc(btuart, btuart->rskb);
					}
					btuart->rskb = NULL;
					btuart->pos = 0;
				}
			}
		}
	}
}


/*
 * Function btuartld_write_wakeup (tty)
 *
 *    Called by the driver when there's room for more data.  If we have
 *    more packets to write, we write them here.
 *
 */
void btuartld_write_wakeup(struct tty_struct *tty)
{
	btuart_device 	*btuart = (btuart_device *) tty->disc_data;
	int 		actual;
	struct sk_buff	*skb;

	DBFENTER;

	/* First make sure we're connected. */
	if (!btuart || btuart->magic != BTUARTLD_MAGIC) {
		return;
	}

	skb = btuart->sskb;
	if (skb == NULL) {
		btuart->tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
		if (!(btuart->hw.flags & AFFIX_UART_RI)) 
			hci_wake_queue(btuart->dev, btuart->flowmask);
		goto exit;
	}

	if (!(btuart->hw.flags & AFFIX_UART_RI) && btuart->hw.proto == HCI_UART_TLP) {
		udelay(BTUART_DELAY);
		btuart_wait_sync(btuart);
	}

	spin_lock_bh(&btuart->lock);	/* Important */

	btuart->tty->flags |= (1 << TTY_DO_WRITE_WAKEUP);
	actual = btuart->tty->driver.write(btuart->tty, 0, skb->data, skb->len);
	DBPRT(DBINIT, "Transmit %d bytes, len was: %d\n", actual, skb->len );

	if (actual == skb->len) {
		kfree_skb(skb);
		btuart->sskb = NULL;
	} else {
		skb_pull(skb, actual);
	}
	
	spin_unlock_bh(&btuart->lock);	/* Important */
 exit:
	DBFEXIT;
}

int btuart_write_generic(btuart_device *btuart, struct sk_buff *skb, int type)
{
	int	actual;

	DBFENTER;

	BTASSERT(btuart, return -1;);

	if (btuart->sskb == NULL) {
		__put_u8(skb_push(skb, 1), type);
		btuart->slen = skb->len;

		DBPRTPKT(DBXMIT, skb->data, skb->len);

		spin_lock_bh(&btuart->lock);	/* Important */
		btuart->tty->flags |= (1 << TTY_DO_WRITE_WAKEUP);
		actual = btuart->tty->driver.write(btuart->tty, 0, skb->data, skb->len);
		DBPRT(DBXMIT, "Transmit %d bytes, len was: %d\n", actual, skb->len );

		if (actual == skb->len) {
			kfree_skb(skb);
		} else {
			skb_pull(skb, actual);
			btuart->sskb = skb;
		}
		spin_unlock_bh(&btuart->lock);	/* Important */
	} else {
		/* busy at the moment */
		return -EBUSY;
	}

	DBFEXIT;
	return 0;
}


int btuart_write_ncc(btuart_device *btuart, struct sk_buff *skb, int type)
{
	int		actual;
	tlp_hdr_t	*tlp;
	__u16		dlen = skb->len;
	struct sk_buff	*s;

	DBFENTER;

	BTASSERT(btuart, return -1;);

	if (btuart->sskb == NULL) {
		if (TLP_PAD_LEN(dlen)) {
			if (skb_tailroom(skb)) {
				*skb_put(skb,1) = 0;
			} else {
				s = dev_alloc_skb(TLP_HDR_LEN + dlen + 1);
				skb_reserve(s, TLP_HDR_LEN);
				memcpy(skb_put(s, dlen), skb->data, dlen);
				*skb_put(s, 1) = 0;	/*PAD*/
				kfree_skb(skb);
				skb = s;
			}
		}
		tlp = (tlp_hdr_t*)skb_push(skb, TLP_HDR_LEN);
		tlp->type = type;
		tlp->zero = 0;
		tlp->length = __htob16(dlen);

		btuart->slen = dlen;

		DBPRT(DBXMIT, "Transmit TLP packet, length: %d\n", skb->len);
		DBPRTPKT(DBXMIT, skb->data, skb->len);

		if (!(btuart->hw.flags & AFFIX_UART_RI)) {
			udelay(BTUART_DELAY);
			btuart_wait_sync(btuart);
		}

		spin_lock_bh(&btuart->lock);	/* Important */
		btuart->tty->flags |= (1 << TTY_DO_WRITE_WAKEUP);
		actual = btuart->tty->driver.write(btuart->tty, 0, skb->data, skb->len);
		DBPRT(DBXMIT, "Transmited %d bytes\n", actual);

		if (actual == skb->len) {
			kfree_skb(skb);
		} else {
			skb_pull(skb, actual);
			btuart->sskb = skb;
		}
		spin_unlock_bh(&btuart->lock);	/* Important */
	} else {
		/* busy at the moment */
		return -EBUSY;
	}

	DBFEXIT;
	return 0;
}

int btuartld_init(btuart_device *btuart)
{
	return btuart->dev_desc.ldisc_init ? btuart->dev_desc.ldisc_init(btuart) : -1;
}

struct file *open_device(kdev_t kdev)
{
	int		err;
	struct file	*file = NULL;
#if 1
	struct inode	*inode;
	mm_segment_t	old_fs;
	struct qstr	qstr;
	char		name[32];

	DBFENTER;
	
	inode = get_empty_inode();
	if (!inode)
		return NULL;

	inode->i_mode = S_IFCHR | S_IRWXUGO;
	inode->i_sock = 0;
	inode->i_uid = current->fsuid;
	inode->i_gid = current->fsgid;
	init_special_inode(inode, inode->i_mode, kdev_t_to_nr(kdev));

	file = get_empty_filp();
	if (file == NULL) {
		iput(inode);
		return NULL;
	}
	
	file->f_op = inode->i_fop;		//!!!
	file->f_mode = inode->i_mode;//3
	file->f_flags = O_RDWR;
	file->f_pos = 0;

	sprintf(name, "[%lu]", inode->i_ino);
	qstr.name = name;
	qstr.len = strlen(name);
	qstr.hash = inode->i_ino;

	file->f_dentry = d_alloc(NULL, &qstr); 
	if (!file->f_dentry) {
		put_filp(file);
		iput(inode);
		return NULL;
	}
	d_add(file->f_dentry, inode);
	old_fs = get_fs(); set_fs(KERNEL_DS);
	err = file->f_op->open(inode, file);//chrdev_open(inode, file);
	set_fs(old_fs);
	if (err) {
		file->f_dentry = NULL;
		put_filp(file);
		iput(inode);
		return NULL;
	}
#else
	char	path[DEVPATHLEN];
	char	*name;

	DBFENTER;

	sprintf(path, "/dev/ttyS%d", minor(kdev)-0x40);
	name = path;

	DBPRT(DBINIT, "Opening device, path: %s\n", name);
	file = filp_open(name, 0, 0);
	if (IS_ERR(file)) {
		BTERRORF("Unable to open divice: %s\n", name);
		return NULL;
	}
#endif
	
	DBFEXIT;
	return file;
}

int close_device(struct file *file)
{
	int		err = 0;
	struct dentry	*dentry = file->f_dentry;
	struct inode	*inode = dentry->d_inode;

	DBFENTER;
#if 0	
	DBPRT(DBINIT, "fcount: %d, icount: %d, dcount: %d\n",
			file_count(file), atomic_read(&inode->i_count),
			atomic_read(&dentry->d_count));
#endif

#if 0
	if (file->f_op->release != NULL) {
		mm_segment_t	old_fs;
		old_fs = get_fs(); set_fs(KERNEL_DS);
		lock_kernel();
		err = file->f_op->release(inode,file);
		unlock_kernel();
		set_fs(old_fs);
	}

	dput(file->f_dentry);
	put_filp(file);
#else
	err = filp_close(file, NULL);
#endif
	iput(inode);

	DBFEXIT;
	return err;

}

#define DEVPATHLEN	256

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 17)
#define devfs_get_handle	devfs_find_handle
#define devfs_put(a)
#endif

struct file *kdev_open(kdev_t kdev)
{
	struct file	*file;
#ifdef CONFIG_DEVFS_FS
	char		path[DEVPATHLEN];
	char		*name;
	devfs_handle_t	handle;
	int		pos;

	DBFENTER;

	handle = devfs_get_handle (NULL, NULL, major(kdev), minor(kdev), DEVFS_SPECIAL_CHR, 0);
	if (handle == NULL) {
		BTERRORF("Unable to find device\n");
		return NULL;
	}
	pos = devfs_generate_path(handle, path, DEVPATHLEN);
	devfs_put(handle);
	if (pos < 0) {
		BTERRORF("Unable to generate a path\n");
		return NULL;
	}
	/* add /dev/ */
	name = path + pos - 5;
	memcpy(name, "/dev/", 5);

	DBPRT(DBINIT, "Opening device, path: %s\n", name);
	file = filp_open(name, 0, 0);
	if (IS_ERR(file)) {
		BTERRORF("Unable to open divice: %s\n", name);
		return NULL;
	}

#else /* no devfs */
	DBFENTER;

	file = open_device(kdev);
#endif

	DBFEXIT;
	return file;
}

int kdev_close(struct file *file)
{
	int	err;
	DBFENTER;
#ifdef CONFIG_DEVFS_FS
	err = filp_close(file, NULL);
#else
	err = close_device(file);
#endif
	DBFEXIT;
	return err;
}

int kdev_ioctl(struct file *filp, int cmd, void *arg)
{
	int	err = -1;
	
	if (filp->f_op && filp->f_op->ioctl) {
		mm_segment_t	old_fs;
		old_fs = get_fs(); set_fs(KERNEL_DS);
		err = filp->f_op->ioctl(filp->f_dentry->d_inode, filp, cmd, (unsigned long)arg);
		set_fs(old_fs);
	}
	return err;
}

static inline int set_latency(btuart_device *btuart, int low)
{
	struct serial_struct	ser;
	int	err;

	DBFENTER;
	
	err = kdev_ioctl(btuart->filp, TIOCGSERIAL, &ser);
	if (err)
		goto ioerr;
	if (low) 
		ser.flags |= ASYNC_LOW_LATENCY;
	else
		ser.flags &= ~ASYNC_LOW_LATENCY;
	err = kdev_ioctl(btuart->filp, TIOCSSERIAL, &ser);
	if (err)
		goto ioerr;

	DBFEXIT;
	return 0;
ioerr:
	return err;	
}

/*
#define    B57600 0010001
#define   B115200 0010002
#define   B230400 0010003
#define   B460800 0010004
#define   B500000 0010005
#define   B576000 0010006
#define   B921600 0010007
*/
static inline int set_speed(btuart_device *btuart, int speed)
{
	struct termios	term;
	int		err;

	DBFENTER;
	DBPRT(DBINIT, "Setting speed: %o\n", speed);

	err = kdev_ioctl(btuart->filp, TCGETS, &term);
	if (err)
		goto ioerr;
	term.c_cflag = (term.c_cflag & ~CBAUD) | speed;
	err = kdev_ioctl(btuart->filp, TCSETS, &term);
	if (err)
		goto ioerr;

	DBFEXIT;
	return 0;
ioerr:
	return err;	
}

static inline int set_ctl(btuart_device *btuart, int ctl)
{
	struct termios	term;
	int		err;

	DBFENTER;
	
	/*  Set crtscts , thus the serial driver will enable modem status
	    change interrupts which we want to monitor.
	*/ 
	err = kdev_ioctl(btuart->filp, TCGETS, &term);
	if (err)
		goto ioerr;
	if (ctl)
		term.c_cflag |= CRTSCTS;
	else
		term.c_cflag &= ~CRTSCTS;
	err = kdev_ioctl(btuart->filp, TCSETS, &term);
	if (err)
		goto ioerr;

ioerr:
	DBFEXIT;
	return err;	
}


int init_uart(btuart_device *btuart)
{
	int	err;

	DBFENTER;

	err = set_latency(btuart, btuart->hw.flags & AFFIX_UART_LOW);
	if (err) {
		BTWARN("Unable to set latency: %d\n", err);
#if 0		
		goto ioerr;
#endif
	}

	err = set_ctl(btuart, btuart->hw.flags & AFFIX_UART_CTL);
	if (err) {
		BTERROR("Unable to set CRTSCTS: %d\n", err);
		goto ioerr;
	}

	err = set_speed(btuart, btuart->hw.speed);
	if (err) {
		BTERROR("Unable to set speed: %d\n", err);
		goto ioerr;
	}

	DBFEXIT;
ioerr:
	return err;
}

int btuartld_init_generic(btuart_device *btuart)
{
	int	disc = N_BTUART, err;

	DBFENTER;

	if (btuart->filp)
		return 0;

	btuart->filp = kdev_open(btuart->hw.kdev);
	if (btuart->filp == NULL)
		return -EBUSY;

	err = init_uart(btuart);
	if (err)
		goto ioerr;
	
	err = kdev_ioctl(btuart->filp, TIOCSETD, &disc);
	if (err)
		goto ioerr;

	DBFEXIT;
	return err;
ioerr:
	DBPRT(DBINIT, "Unable to set discipline/terminal parameters: err: %d\n", err);
	kdev_close(btuart->filp);
	btuart->filp = NULL;
	return -EBUSY;
}


static inline int enable_RI(btuart_device *btuart)
{
	int	err;

	DBFENTER;
	
	if (!(btuart->hw.flags & AFFIX_UART_RI))
		return 0;

	/*  Set crtscts , thus the serial driver will enable modem status
	    change interrupts which we want to monitor.
	*/ 
	err = set_ctl(btuart, 1);
	if (err) {
		BTERROR("Unable to set CRTSCTS: %d\n", err);
		goto ioerr;
	}

	/*  Read initial value of RI bit */
	btuart->uart_ri_latch = inb(btuart->hw.port + UART_MSR) & UART_MSR_RI;
ioerr:
	DBFEXIT;
	return err;	
}

int btuartld_init_ncc(btuart_device *btuart)
{
	int	disc = N_BTUART, err;

	DBFENTER;

	if (btuart->filp) 
		return enable_RI(btuart);

	btuart->filp = kdev_open(btuart->hw.kdev);
	if (btuart->filp == NULL) 
		return -EBUSY;

	err = init_uart(btuart);
	if (err)
		goto ioerr;

	err = kdev_ioctl(btuart->filp, TIOCSETD, &disc);
	if (err)
		goto ioerr;

	/* check if port & irq is zero -> get it */
	if (btuart->hw.port == 0 && btuart->hw.irq == 0) {
		struct serial_struct	tmp;
		err = kdev_ioctl(btuart->filp, TIOCGSERIAL, &tmp);
		btuart->hw.port = tmp.port;
		btuart->hw.irq = tmp.irq;
	}
	
	err = enable_RI(btuart);
	if (err)
		goto ioerr;
	/*  Setup an interrupt service routine to handle the "RI" flow control.
	    This type of flow control is Nokia specific.
	    Modem Status Change interrupts are enabled by the serial driver
	*/
	if (btuart->hw.flags & AFFIX_UART_RI) {
		err = request_irq(btuart->hw.irq, btuart_isr, SA_SHIRQ, BTUART_DEVNAME, btuart);
		if (err)
			goto ioerr;
	}

	DBFEXIT;
	return 0;
ioerr:
	DBPRT(DBINIT, "Unable to set discipline/terminal parameters: err: %d\n", err);
	kdev_close(btuart->filp);
	btuart->filp = NULL;
	return -EBUSY;
}

void btuartld_uninit(btuart_device *btuart)
{
	if (btuart->dev_desc.ldisc_uninit)
		btuart->dev_desc.ldisc_uninit(btuart);
}

void btuartld_uninit_generic(btuart_device *btuart)
{
	DBFENTER;

	if (btuart->filp == NULL)
		return;
	
	kdev_close(btuart->filp);
	btuart->filp = NULL;

	DBFEXIT;
}

void btuartld_uninit_ncc(btuart_device *btuart)
{
	DBFENTER;

	if (btuart->filp == NULL)
		return;
	
	if (btuart->hw.flags & AFFIX_UART_RI)
		free_irq(btuart->hw.irq, btuart);

	kdev_close(btuart->filp);
	btuart->filp = NULL;

	DBFEXIT;
}

/***********************   Network Interface Subsystem  ********************/

int btuart_open(struct net_device *dev)
{
	int		err;
	btuart_device	*btuart = dev->priv;

	DBFENTER;

	err = btuartld_init(btuart);
	if (err)
		return err;

	btuart->flowmask = HCI_CMD_ON | HCI_ACL_ON | HCI_SCO_ON;

	if (btuart->hw.count)
		(*btuart->hw.count)++;

	hci_start_queue(btuart->dev, btuart->flowmask);

	DBFEXIT;
	return 0;
}

int btuart_stop(struct net_device *dev)
{
	btuart_device	*btuart = dev->priv;
	DBFENTER;

	hci_stop_queue(btuart->dev, btuart->flowmask);

	btuartld_uninit(btuart);
	if (btuart->hw.count)
		(*btuart->hw.count)--;

	DBFEXIT;
	return 0;
}

int btuart_ioctl(struct net_device *dev, void *arg, int cmd)
{
	DBFENTER;

	switch (cmd) {

	default:
		return -ENOIOCTLCMD;
	}
	
	DBFEXIT;
	return 0;
}

struct net_device_stats * btuart_get_stats(struct net_device *dev)
{
	btuart_device	*btuart = dev->priv;

	return &btuart->stats;
}

int btuart_xmit_generic(struct sk_buff *skb, struct net_device *dev)
{
	btuart_device	*btuart = dev->priv;
	hci_skb_cb	*cb = (hci_skb_cb*)skb->cb;
	int		type;

	DBFENTER;

	if (!hci_running(dev)) {
		DBPRT(DBXMIT, "%s: xmit call when iface is down\n", dev->name);
		dev_kfree_skb(skb);
		return 0;
	}

	if( hci_queue_stopped(dev)) {
		DBPRT(DBXMIT, "Busy\n");
		return -EBUSY;
	}

	if (skb == NULL) {
		DBPRT(DBXMIT, "We are ready to stransmit\n");
		//dev_tint(dev);
		return 0;
	}

	dev->trans_start = jiffies;
	skb->dev = dev;
	
	/* check flow now */
	switch (cb->type) {
	case HCI_COMMAND:
		type = UART_HCI_COMMAND;
		break;

	case HCI_ACL:
		type = UART_HCI_ACL;
		break;

	case HCI_SCO:
		type = UART_HCI_SCO;
		break;

	default:
		type = 0;
		DBPRT(DBXMIT, "Unknown type of packet: %d\n", type);
	}

	hci_stop_queue(dev, btuart->flowmask);

	return btuart_write_generic(btuart, skb, type);
}

int btuart_xmit_ncc(struct sk_buff *skb, struct net_device *dev)
{
	btuart_device	*btuart = dev->priv;
	hci_skb_cb	*cb = (hci_skb_cb*)skb->cb;
	int		type;

	DBFENTER;

	if (!hci_running(dev)) {
		DBPRT(DBXMIT, "%s: xmit call when iface is down\n", dev->name);
		dev_kfree_skb(skb);
		return 0;
	}

	if( hci_queue_stopped(dev)) {
		DBPRT(DBXMIT, "Busy\n");
		return -EBUSY;
	}

	if (skb == NULL) {
		DBPRT(DBXMIT, "We are ready to stransmit\n");
		//dev_tint(dev);
		return 0;
	}

	dev->trans_start = jiffies;
	skb->dev = dev;
	
	/* check flow now */
	switch (cb->type) {
	case HCI_COMMAND:
		type = TLP_HCI_COMMAND;
		break;

	case HCI_ACL:
		type = TLP_HCI_ACL;
		break;

	case HCI_SCO:
		type = TLP_HCI_SCO;
		break;

	default:
		type = 0;
		DBPRT(DBXMIT, "Unknown type of packet: %d\n", type);
	}
	/*  Stop the sending queue. It will be reenabled in btuart_isr
	    when the RI bit of the UART Modem Status Register will change.
	*/
	hci_stop_queue(dev, btuart->flowmask);

	return btuart_write_ncc(btuart, skb, type);
}

int btuart_init(struct net_device *dev)
{
	DBFENTER;


	DBFEXIT;
	return 0;
}

void btuart_uninit(struct net_device *dev)
{
	DBFENTER;


	DBFEXIT;
}

/*
  USE
  dev = init_netdev()
*/
int btuart_register_netdev(btuart_device *btuart)
{
	struct net_device	*dev;
	int	err;

	DBFENTER;	
	dev = affix_alloc_btdev();
	if (dev == NULL)
		return -ENOMEM;

	dev->priv = btuart;	/* set private pointer */
	dev->init = btuart_init;
	dev->uninit = btuart_uninit;
	dev->open = btuart_open;
	dev->stop = btuart_stop;
	dev->do_ioctl = (void*)btuart_ioctl;
	dev->get_stats = btuart_get_stats;
	dev->hard_start_xmit = btuart->dev_desc.net_xmit;
	dev->hard_header_len = btuart->dev_desc.hard_hdr_len;

	btuart->dev = dev;

	SET_MODULE_OWNER(dev);

	rtnl_lock();
	err = register_netdevice(dev);
	rtnl_unlock();
	DBFEXIT;
	return err;
}

void btuart_unregister_netdev(btuart_device *btuart)
{
	DBFENTER;
	rtnl_lock();
	dev_close(btuart->dev);
	unregister_netdevice(btuart->dev);
	rtnl_unlock();
	btuart->dev = NULL;
	DBFEXIT;
}

static btuart_dev_desc_t *get_by_id(int manfid, int prodid, int proto)
{
	int	i;
	int	m_id, p_id, p;
	
	if (manfid == 0 && prodid == 0)
		return NULL;
	for (i = 0; i < NOF_UART_DEVS; i++) {
		m_id = btuart_dev_desc[i].manfid;
		p_id = btuart_dev_desc[i].prodid;
		p = btuart_dev_desc[i].proto;
		
		if ((manfid == m_id && prodid == p_id) 
				&& ((proto == 0) || (proto == p)))
			return &btuart_dev_desc[i];
	}
	return NULL;
}

static btuart_dev_desc_t *get_by_type(int type, int proto)
{
	int	i;
	int	t, p;
	
	if (type == -1)
		return NULL;
	for (i = 0; i < NOF_UART_DEVS; i++) {
		t = btuart_dev_desc[i].type;
		p = btuart_dev_desc[i].proto;
		
		if (type == t && ((proto == 0) || (proto == p)))
			return &btuart_dev_desc[i];
	}
	return NULL;
}

static btuart_dev_desc_t *get_by_proto(int proto)
{
	int	i;
	int	p;
	
	if (proto == 0)
		return NULL;
	for (i = 0; i < NOF_UART_DEVS; i++) {
		p = btuart_dev_desc[i].proto;
		
		if (proto == p)
			return &btuart_dev_desc[i];
	}
	return NULL;
}



int affix_uart_attach(affix_uart_t *serial)
{
	int	err;
	btuart_device		*btuart;
	btuart_dev_desc_t	*dd;
	
	DBFENTER;

	if (!try_inc_mod_count(serial->owner))
		return -ENODEV;

	btuart = btuart_lookup_device(serial->kdev);
	if (btuart) {
		err = -EBUSY;
		goto err;
	}

	btuart = btuart_create();
	if (!btuart) {
		err = -ENOMEM;
		goto err;
	}

	/* select configuration */
	dd = get_by_id(serial->manfid, serial->prodid, serial->proto);
	if (!dd) {
		dd = get_by_type(serial->type, serial->proto);
		if (!dd) {
			dd = get_by_proto(serial->proto);
			if (!dd) {
				// try default
				dd = &btuart_dev_desc[NOF_UART_DEVS-1];
			}
		}
	}

	btuart->dev_desc = *dd;
	btuart->hw = *serial;
	/* proto <> desc */ 
	btuart->hw.proto = btuart->dev_desc.proto;
	if (!btuart->hw.speed)	/* set default speed */
		btuart->hw.speed = btuart->dev_desc.speed;
	if (!btuart->hw.flags)
		btuart->hw.flags = btuart->dev_desc.flags;
	if (btuart->hw.type == -1 )
		btuart->hw.type = btuart->dev_desc.type;
	if (btuart->hw.manfid == 0 && btuart->hw.prodid == 0) {
		btuart->hw.manfid = btuart->dev_desc.manfid;
		btuart->hw.prodid = btuart->dev_desc.prodid;
	}

	tasklet_init(&btuart->rx_task, (void*)btuart->dev_desc.defer_receive_buf, (unsigned long) btuart);

	err = btuart_register_netdev(btuart);
	if (!err)
		return 0;
err:
	if (serial->owner)
		__MOD_DEC_USE_COUNT(btuart->hw.owner);
	return err;
}

int affix_uart_detach(kdev_t kdev)
{
	btuart_device *btuart;

	DBFENTER;

	btuart = btuart_lookup_device(kdev);
	if (btuart == NULL)
		return -ENODEV;

	if (btuart->dev)
		btuart_unregister_netdev(btuart);
	btuartld_uninit(btuart);
	btuart->hw.count = NULL;

	if (btuart->hw.owner)
		__MOD_DEC_USE_COUNT(btuart->hw.owner);

	btuart_destroy(btuart);

	DBFEXIT;
	return 0;
}

void affix_uart_suspend(kdev_t kdev)
{
	btuart_device *btuart;

	DBFENTER;

	btuart = btuart_lookup_device(kdev);
	if (btuart == NULL)
		return;

	if (btuart->dev) {
		hci_device_detach(btuart->dev);
		//btuartld_uninit(btuart);
	}

	DBFEXIT;
	return;
}

void affix_uart_resume(kdev_t kdev)
{
	btuart_device *btuart;

	DBFENTER;

	btuart = btuart_lookup_device(kdev);
	if (btuart == NULL)
		return;

	if (btuart->dev) {
		btuartld_init(btuart);
		hci_device_attach(btuart->dev);
	}
	DBFEXIT;
	return;
}

struct affix_uart_operations btuart_ops = {
	owner: THIS_MODULE,
	attach: affix_uart_attach,
   	detach: affix_uart_detach,
	suspend: affix_uart_suspend
};
	 
/*
     Function:
     btuart_init register line discipline for serial tty
*/
int __init init_btuart(void)
{
	int	err;

	DBFENTER;

	btl_head_init(&btuarts);
	/* Fill in our line protocol discipline, and register it */
	memset(&btuart_ldisc, 0, sizeof(btuart_ldisc));

	btuart_ldisc.magic = TTY_LDISC_MAGIC;
	btuart_ldisc.name  = "n_affix";
	btuart_ldisc.flags = 0;
	btuart_ldisc.open  = btuartld_open;
	btuart_ldisc.close = btuartld_close;
	btuart_ldisc.read  = NULL;
	btuart_ldisc.write = NULL;
	btuart_ldisc.ioctl = (int (*)(struct tty_struct *, struct file *,
				      unsigned int, unsigned long)) btuartld_ioctl;
	btuart_ldisc.poll  = NULL;
	btuart_ldisc.receive_buf  = btuartld_receive_buf;
	btuart_ldisc.receive_room = btuartld_receive_room;
	btuart_ldisc.write_wakeup = btuartld_write_wakeup;
	
	if (( err = tty_register_ldisc(N_BTUART, &btuart_ldisc)) != 0) {
		BTERRORF("Can't register line discipline (err = %d)\n", err);
	}
	
	if (err)
		goto exit;
	
	err = affix_set_uart(&btuart_ops);

exit:	
	DBFEXIT;
	return err;
}

void __exit exit_btuart(void)
{
	int err;
  
	DBFENTER;

	affix_set_uart(NULL);

	if ((err = tty_register_ldisc(N_BTUART, NULL)))	{
		BTERROR("can't unregister line discipline (err = %d)\n", err);
	}

	DBFEXIT;
}


/*  If we are resident in kernel we want to call init_btuart_cs manually. */
#ifdef MODULE
module_init(init_btuart);
module_exit(exit_btuart);

MODULE_AUTHOR("Dmitry Kasatkin <dmitry.kasatkin@nokia.com>");
MODULE_DESCRIPTION("BTUARTLD driver for UART Bluetooth cards");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 15)
MODULE_LICENSE("GPL");
#endif

#endif // MODULE

EXPORT_SYMBOL(affix_uart_attach);
EXPORT_SYMBOL(affix_uart_detach);
EXPORT_SYMBOL(affix_uart_suspend);
EXPORT_SYMBOL(affix_uart_resume);


/*
 * TIOCSSERIAL TIOCGSERIAL 
 * struct serial struct
 */


/* 
 * issues
 * 1. CRTSCTS
 * 2. speed
 * 3. latency
 *
 */
