/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btuart.h,v 1.2 2002/09/03 09:15:09 kds Exp $

   BTUART - physical protocol layer for bluetooth uart devices

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/

#ifndef _BTUART_H
#define _BTUART_H

#define	BTUARTLD_MAGIC	0x94765765
#define BTUART_DEVNAME	"affix_uart"

#include <linux/skbuff.h>
#include <linux/netdevice.h>

#include <affix/types.h>
#include <affix/hci.h>

#define N_BTUART	14	/* line discipline */

/* NOKIA */
#define MANFID_NOKIA		0x0124
#define PRODID_DTL1_PROTO	0x0900
#define PRODID_DTL1		0x1000

/* SOCKET */
#define MANFID_SOCKET		0x0104
#define PRODID_SOCKET_BTCF1	0x009f



/* support for the Nokia Connectivity Card */

/* the structure of the TLP protocol header */
typedef struct {
	__u8	type;
	__u8	zero;
	__u16	length;
} __PACK__ tlp_hdr_t;

#define TLP_HDR_LEN		4

#define TLP_CONTROL		0x80
#define TLP_HCI_COMMAND		0x81
#define TLP_HCI_ACL		0x82
#define TLP_HCI_SCO		0x83
#define TLP_HCI_EVENT		0x84

#define TLP_PAD_LEN(len)	(len & 0x0001)
#define TLP_FCS_LEN		2

#define NCC_SYNC	0x80
#define NCC_SYNC_TIME	(5000)


#define UART_HCI_COMMAND	1
#define UART_HCI_ACL		2
#define UART_HCI_SCO		3
#define UART_HCI_EVENT		4

struct btuart_device;

typedef struct {
	int	proto;
	int	flags;
	int	speed;	/* initial speed (may not work in other case) */
	int	type;
	int	manfid;
	int	prodid;
	int	hard_hdr_len;
	int	(*net_xmit)(struct sk_buff *skb, struct net_device *dev);
	int	(*ldisc_init)(struct btuart_device *btuart);
	void	(*ldisc_uninit)(struct btuart_device *btuart);
	void	(*ldisc_receive_buf)(struct tty_struct *, const unsigned char *, char *, int);
	void	(*defer_receive_buf)(void *data);
} btuart_dev_desc_t;

/* control structure for transport driver */
typedef struct btuart_device {
	struct list_head	q;	/* for queueing */

	int			magic;

	/* physical port stuff */
	affix_uart_t		hw;
	btuart_dev_desc_t	dev_desc;

	/* line discipline stuff */
	struct file		*filp;		/* for tty driver */
	struct tty_struct 	*tty;
	spinlock_t		lock;

	unsigned long		flowmask;	/* hci flow mask */
	int			sync;
	struct timeval		sync_stamp;

	struct sk_buff		*sskb;	/* sended packet */
	__u32			slen;

	/* receiver state for generic uart mode */
	int			pos;
	int			plen;
	tlp_hdr_t		tlp;
	struct sk_buff		*rskb;
	int			hlen;
	int			rpos;
	__u8			hdr[5];		/* 1 (type) + 4 (ACL) */
	
	
	/* for fast irq handling */
	struct sk_buff_head	rx_queue;
	struct tasklet_struct	rx_task;

	/* network device subsystem */
	struct net_device_stats	stats;
	struct net_device	*dev;

	int			uart_ri_latch;
#ifdef CONFIG_AFFIX_DEBUG
	int			irqcnt;
#endif
} btuart_device;

void btuart_receive_buf(void *data);
int btuartld_init(btuart_device *btuart);
void btuartld_uninit(btuart_device *btuart);

int affix_uart_attach(affix_uart_t *serial);
int affix_uart_detach(kdev_t kdev);
void affix_uart_suspend(kdev_t kdev);
void affix_uart_resume(kdev_t kdev);

#endif
