/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/* 
   $Id: bluetooth.h,v 1.14 2002/09/03 07:34:36 kds Exp $

   AF_L2CAP - L2CAP Address family for socket interface

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#ifndef	_AF_AFFIX_H
#define	_AF_AFFIX_H

#ifdef __KERNEL__
#include <linux/if.h>
#include <linux/ioctl.h>
#else
#include <net/if.h>
#include <stdint.h>
#endif

#include <linux/socket.h>

#include <affix/types.h>
#include <affix/hci_types.h>

/* kernel only section */
#ifdef __KERNEL__

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,18)
#define mk_kdev(a,b) MKDEV(a,b)
#define major(d) MAJOR(d)
#define minor(d) MINOR(d)
#define kdev_same(a,b) (a==b)
#define kdev_none(d) (!(d))
#endif

/*
 * error codes
 */
#define ERR_OK			0x0000
#define ERR_PENDING		0x0001	// pending...				EINPROGRESS
#define ERR_LINKTO		0x0002	// baseband link timed out		EHOSTUNREACH
#define ERR_RESOURCE		0x0003	// no resources - 			ENOMEM
#define ERR_CONNREFUSED		0x0003	// service not available (on PSM)	ECONNREFUSED
#define ERR_SECURITY		0x0004	// access not allowed	 		EACCESS
#define ERR_TIMEDOUT		0x0005	// connection timed out			ETIMEDOUT
#define ERR_CONNRESET		0x0006	// disconnection came from peer side	ECONNRESET
#define ERR_FAILURE		0x0007	// any failure
#define ERR_INVALIDARGS		0x0008	// invalid arguments (proto specific)	EINVAL
#define ERR_HARDWARE		0x0009	// hardware failure

/*
 * general connection states
 */
typedef volatile enum {
	DEAD = 0,
	CON_OPEN = 1,
	CON_CLOSED,
	CON_W4_LINKUP,		/* wait for link is up */
	CON_W4_LCONREQ,
	CON_W4_CONRSP,
	CON_W4_CONREQ,
	CON_W4_LCONRSP,
	CON_CONFIG,		/* protocol specific configuration process */
	CON_W4_LDISCREQ,
	CON_W4_DISCRSP,
	CON_W4_DISCREQ,
	CON_W4_LDISCRSP,
	CON_AUTH_PENDING,

	/* socket specific */
	CON_ESTABLISHED = CON_OPEN,
	CON_CONNECTING = CON_W4_CONRSP,
	CON_LISTEN = CON_W4_CONREQ,
	
	CON_MAX_STATES /* Leave at the end! */
} con_state;


int affix_sock_register(struct net_proto_family *pf, int protocol);
int affix_sock_unregister(int protocol);

#endif	/* __KERNEL__ */



#define FLOW_STOP	0x00
#define FLOW_START	0x01

/* IOCTL for HCI devices */
#define SIOCHCI_SET_AUDIO	(SIOCDEVPRIVATE+1)
#define SIOCHCI_GET_AUDIO	(SIOCDEVPRIVATE+2)

/* Bypass warnings */
#ifdef AF_AFFIX
#undef AF_AFFIX
#endif

#ifdef PF_AFFIX
#undef PF_AFFIX
#endif

#define AF_AFFIX	27		// FIXME:
#define PF_AFFIX	AF_AFFIX

/* Socket Options Stuff */
#define SOL_AFFIX	277

#define BTSO_HCI	0x00
#define BTSO_PIN	0x01
#define BTSO_MTU	0x02
/* security */
#define BTSO_SECLEVEL	0x03
#define SVCSEC_AUTHOR	0x01
#define SVCSEC_AUTH	0x02
#define SVCSEC_ENCRYPT	0x04

/* network device definition */
#define ARPHRD_BLUETOOTH	0x8000
#define ETH_P_BLUETOOTH		0x0027

#define BTPROTO_HCI	0
#define BTPROTO_L2CAP	1
#define BTPROTO_RFCOMM	2
#define BTPROTO_HCIACL	3
#define BTPROTO_HCISCO	4
#define	BTPROTO_MAX	10	/* should be 1 greater then last proto */

/* HCI address family data types */

//#define SIOCAFFIX_BINDHCI	(SIOCPROTOPRIVATE+8)	/* common IOCTL to all sockets */

struct sockaddr_affix {
	sa_family_t	family;
	BD_ADDR		bda;
	uint16_t	port;
	BD_ADDR		local;		/* local Bluetooth adapter to connect to */
};


/* L2CAP address family data types */

/* cmsg_type */
#define L2CAP_PING		0
#define L2CAP_SINGLEPING	1

//struct l2cap_pkt_info {
//	int
//}

/* sioc commands for socket interface */
#define SIOCL2CAP_CONFIG	(SIOCPROTOPRIVATE+0)
#define SIOCL2CAP_FLUSH		(SIOCPROTOPRIVATE+1)


/* RFCOMM stuff */
#define SIOCRFCOMM_BINDPORT		(SIOCPROTOPRIVATE+0)
#define SIOCRFCOMM_DISCONNECTPORT	(SIOCPROTOPRIVATE+1)
#define SIOCRFCOMM_CHECKPORTS		(SIOCPROTOPRIVATE+2)
#define SIOCRFCOMM_SETTYPE		(SIOCPROTOPRIVATE+4)

struct rfcomm_port {
	int		line;
	BD_ADDR		bda;
	int		channel;
};

struct rfcomm_ports {
	struct rfcomm_port	*info;	/* array ptr */
	int			size;	/* array size */
	int			count;	/* read info */
};

#define RFCOMM_SOCKET	0
#define RFCOMM_BTY	1


/* PAN stuff */
#define SIOCSFILTERPROTOCOL	SIOCDEVPRIVATE		/* set protocol filter */
#define SIOCSFILTERMULTICAST	SIOCDEVPRIVATE + 1	/* set mulitcast filter */
#define SIOCGFILTERPROTOCOL	SIOCDEVPRIVATE + 2	/* get protocol filter */
#define SIOCGFILTERMULTICAST	SIOCDEVPRIVATE + 3	/* get mulitcast filter */


#define PROTOCOL_FILTER_MAX	16
#define MULTICAST_FILTER_MAX	16

#define F_START 0
#define F_STOP  1

#ifndef ETH_ALEN
#define ETH_ALEN 6
#endif

typedef unsigned char ETH_ADDR[ETH_ALEN];

typedef struct {
	__u16	count; /* number of filter entrys, 0 = no filter */
	__u16	protocol[PROTOCOL_FILTER_MAX][2]; /* start_range - stop_range pairs in network byte order */
} protocol_filter;

typedef struct {
	__u16		count; /* number of filter entrys, 0 = no filter */
	ETH_ADDR	multicast[MULTICAST_FILTER_MAX][2]; /* start_range - stop_range pairs in network byte order */
} multicast_filter;




/* **** */
#define EVENT_MAX_SIZE	257
#define CMD_MAX_SIZE	258
#define MSG_MAX_SIZE	0x20

struct hci_msg_hdr {
	int	opcode;
	int	length;
};

#define MGR_INIT_HCI		0x01
#define MGR_CONNECT_REQ		0x02
#define MGR_DISCONNECT_REQ	0x03
#define MGR_AUTH_REQ		0x04

struct hci_init_req {
	struct hci_msg_hdr	hdr;
	int			hd;
};

struct hci_connect_req {
	struct hci_msg_hdr	hdr;
	int			id;
};

struct hci_disconnect_req {
	struct hci_msg_hdr	hdr;
	int			id;
	__u8			reason;
};

struct hci_auth_req {
	struct hci_msg_hdr	hdr;
	int			id;
};


/* ioctl stuff */
#define BTIOC_MAGIC	'b'

/* Command exec flags */
#define HCI_W4_STATUS			0x01
#define HCI_W4_COMPLETE			HCI_W4_STATUS
#define HCI_W4_EXTRA_EVENT		0x02
#define HCI_USE_STOP_EVENT		0x04
#define HCI_AUTO_RELEASE		0x08

#define HCI_COMMAND_PENDING		0x80

struct hci_cmd_arg {
	HCI_Command_Packet_Header	*cmd;
	unsigned long			flags;
	__u64				event_mask;	/* waiting */
	__u8				extra_event;
	__u8				stop_event;
	HCI_Event_Packet_Header		*event;
	int				status;		/* Command Status */
};

#define BTIOC_EXEC	_IOWR(BTIOC_MAGIC, 0, struct hci_cmd_arg)
#define BTIOC_LISTEN	_IOW(BTIOC_MAGIC, 1, __u64)

#define HCI_OPEN_ANY	0
#define HCI_OPEN_NAME	1
#define HCI_OPEN_ID	2
#define HCI_OPEN_MGR	3

#define HCI_OPEN_KERNEL	0x80

struct hci_open {
	int	cmd;
	char	name[IFNAMSIZ];
	int	hd;
};

#define BTIOC_OPEN_HCI	_IOW(BTIOC_MAGIC, 2, struct hci_open)

struct hci_event_arg {
	int				status;
	int				hd;
	HCI_Event_Packet_Header		*event;
};

#define BTIOC_RECV_EVENT	_IOWR(BTIOC_MAGIC, 3, struct hci_event_arg)
#define BTIOC_RECV_MSG		_IO(BTIOC_MAGIC, 4)

#define BTIOC_DBMGET	_IOR(BTIOC_MAGIC, 5, __u32)
#define BTIOC_DBMSET	_IOW(BTIOC_MAGIC, 6, __u32)

#define BTIOC_ADDPINCODE	_IOW(BTIOC_MAGIC, 9, struct PIN_Code)
#define BTIOC_REMOVEPINCODE	_IO(BTIOC_MAGIC, 10)
#define BTIOC_REMOVELINKKEY	_IO(BTIOC_MAGIC, 11)

/* mask to enable TLP protocol on UART */
//#define AFFIX_UART_TLP		0x01
#define AFFIX_UART_RI		0x02
#define AFFIX_UART_CTL		0x04
#define AFFIX_UART_LOW		0x08

struct open_uart {
	int	dev;
	int	proto;
	int	flags;	
	int	speed;
	int	type;
};

#define BTIOC_OPEN_UART		_IOWR(BTIOC_MAGIC, 14, struct open_uart)
#define BTIOC_CLOSE_UART	_IOWR(BTIOC_MAGIC, 15, struct open_uart)

#define MAX_DEVS		16
#define BTIOC_GETDEVS		_IOWR(BTIOC_MAGIC, 16, int[MAX_DEVS])

struct affix_audio {
	int	mode;
	__u16	setting;
};

struct hci_conn_info {
	BD_ADDR		bda;
	__u16		handle;
};

#define BTIOC_GET_CONN	_IOWR(BTIOC_MAGIC, 22, struct hci_conn_info)


#define HCI_ATTR_ALL		0xFF
#define HCI_ATTR_PKT_TYPE	0x01
#define HCI_ATTR_ALLOW_SWITCH	0x02
#define HCI_ATTR_ROLE		0x04
#define HCI_ATTR_SECMODE	0x08

#define SECMODE_NONSECURE	0x01
#define SECMODE_SVCLEVEL	0x02
#define SECMODE_LINKLEVEL	0x03

struct hci_dev_attr {
	int		hd;
	int		attr;
	char		name[IFNAMSIZ];
	BD_ADDR		bda;
	__u16		pkt_type;	/* packet type at connection */
	__u8		allow_switch;	/* connection */
	__u8		role;		/* role at connection */
	int		sec_mode;	/* security mode */
};

#define BTIOC_SET_HCI	_IOW(BTIOC_MAGIC, 23, struct hci_dev_attr)
#define BTIOC_GET_HCI	_IOWR(BTIOC_MAGIC, 24, struct hci_dev_attr)

#define AFFIX_MODE_TYPE		0x000F
#define AFFIX_MODE_HCI		0x0001
#define AFFIX_MODE_MGR		0x0002
#define AFFIX_MODE_SCO		0x0004
#define AFFIX_MODE_KERNEL	0x0080
#define AFFIX_MODE_MASK		0xFF00
#define AFFIX_MODE_PIN		0x0100

#define CTL_ATTR_MODE		0x01

struct affix_ctl_attr {
	int	attr;
	int	mode;
};

#define BTIOC_SET_CTL		_IOW(BTIOC_MAGIC, 25, struct affix_ctl_attr)
#define BTIOC_GET_CTL		_IOWR(BTIOC_MAGIC, 26, struct affix_ctl_attr)


#define AFFIX_AUDIO_IN		0x01
#define AFFIX_AUDIO_OUT		0x02

#define BTIOC_SET_AUDIO		_IOW(BTIOC_MAGIC, 27, struct affix_audio)
#define BTIOC_GET_AUDIO		_IOWR(BTIOC_MAGIC, 28, struct affix_audio)

struct affix_version {
	int	version;
};

#define BTIOC_GET_VERSION	_IOR(BTIOC_MAGIC, 29, struct affix_version)


struct affix_sec {
	BD_ADDR		bda;
	int		attr;	/* what is valid */

};

#define BTIOC_SET_SEC		_IOW(BTIOC_MAGIC, 20, struct affix_sec)
#define BTIOC_GET_SEC		_IOWR(BTIOC_MAGIC, 31, struct affix_sec)


/* 
 * HCI stuff
 */
int hci_open(char *name);
int hci_open_id(int hd);
int hci_open_mgr(void);
int __hci_open(char *name);
int __hci_open_id(int hd);
int __hci_open_mgr(void);



// HCI commands - from the spec

// Link Control
// __HCI_xx - unwaitable version of the command. returns after command status event
//
int HCI_Inquiry(int fd, __u8 Inquiry_Length, __u8 Max_Num_Responses,
		INQUIRY_ITEM *Items, __u8 *Num_Responses);
int HCI_InquiryCancel(int fd);
int HCI_PeriodicInquiryMode(int fd, __u16 Max_Period_Length, __u16 Min_Period_Length, 
			    __u16 Inquiry_Length, __u8 Max_Num_Responses);
int HCI_ExitPeriodicInquiryMode(int fd);
int HCI_CreateConnection(int fd, INQUIRY_ITEM *dev, __u16 Packet_Type, __u8 Allow_Role_Switch,
			 CHANDLE *Connection_Handle, __u8 *Link_Type, __u8 *Encryption_Mode);
int __HCI_CreateConnection(int fd, INQUIRY_ITEM *dev, __u16 Packet_Type, __u8 Allow_Role_Switch);
int HCI_Disconnect(int fd, CHANDLE Connection_Handle, __u8 Reason);
int __HCI_Disconnect(int fd, CHANDLE Connection_Handle, __u8 Reason);
int __HCI_AddSCOConnection(int fd, CHANDLE Connection_Handle, __u16 Packet_Type);
int HCI_AcceptConnectionRequest(int fd, BD_ADDR *bda, __u8 Role, CHANDLE *Connection_Handle,
				__u8 *Link_Type, __u8 *Encryption_Mode);
int __HCI_AcceptConnectionRequest(int fd, BD_ADDR *bda, __u8 Role);
int HCI_RejectConnectionRequest(int fd, BD_ADDR *bda, __u8 Reason);
int __HCI_RejectConnectionRequest(int fd, BD_ADDR *bda, __u8 Reason);
int HCI_LinkKeyRequestReply(int fd, BD_ADDR *bda, __u8 *Link_Key);
int HCI_LinkKeyRequestNegativeReply(int fd, BD_ADDR *bda);
int HCI_PINCodeRequestReply(int fd, BD_ADDR *bda, __u8 PIN_Code_Length, __u8 *PIN_Code);
int HCI_PINCodeRequestNegativeReply(int fd, BD_ADDR *bda);
int HCI_ChangeConnectionPacketType(int fd, CHANDLE Connection_Handle, __u16 Packet_Type);
int HCI_AuthenticationRequested(int fd, CHANDLE Connection_Handle);
int __HCI_AuthenticationRequested(int fd, CHANDLE Connection_Handle);
int HCI_SetConnectionEncryption(int fd, CHANDLE Connection_Handle, __u8 Encryption_Enable);
int HCI_ChangeConnectionLinkKey(int fd, CHANDLE Connection_Handle,
				BD_ADDR *bda, __u8 *Link_Key, __u8 *Key_Type);
int HCI_MasterLinkKey(int fd, __u8 Key_Flag, CHANDLE *Connection_Handle);
int HCI_RemoteNameRequest(int fd, INQUIRY_ITEM *dev, char *Name);

// Link Policy Commands
int HCI_ParkMode(int fd, CHANDLE Connection_Handle, __u16 Beacon_Max_Interval, __u16 Beacon_Min_Interval);
int HCI_ExitParkMode(int fd, CHANDLE Connection_Handle, __u8 *Current_Mode, __u16 *Interval);
int HCI_RoleDiscovery(int fd, CHANDLE Connection_Handle, __u8 *Current_Role);
int HCI_SwitchRole(int fd, BD_ADDR *bda, __u8 Role);

// HC & BB commands
int HCI_SetEventMask(int fd, unsigned long long mask);
int HCI_Reset(int fd);
int HCI_SetEventFilter(int fd, __u8 Filter_Type, __u8 Filter_Condition_Type, __u8 *Condition, __u8 Condition_Length);
int HCI_ReadPINType(int fd, __u8 *PIN_Type);
int HCI_WritePINType(int fd, __u8 PIN_Type);
int HCI_CreateNewUnitKey(int fd);
int HCI_ReadStoredLinkKey(int fd, BD_ADDR *bda, __u8 Read_All_Flag, __u16 *Max_Num_Keys, __u16 *Num_Keys, struct Link_Key *Link_Keys);
int HCI_WriteStoredLinkKey(int fd, __u8 Num_Keys_To_Write, struct Link_Key *Link_Keys, __u8 *Num_Keys_Written);
int HCI_DeleteStoredLinkKey(int fd, BD_ADDR *bda, __u8 Delete_All_Flag, __u16 *Num_Keys_Deleted);
int HCI_ChangeLocalName(int fd, char *Name);
int HCI_ReadLocalName(int fd, char *Name);
int HCI_ReadPageTimeout(int fd, __u16 *Page_Timeout);
int HCI_WritePageTimeout(int fd, __u16 Page_Timeout);
int HCI_ReadScanEnable(int fd, __u8 *Scan_Enable);
int HCI_WriteScanEnable(int fd, __u8 Scan_Enable);
int HCI_ReadAuthenticationEnable(int fd, __u8 *Authentication_Enable);
int HCI_WriteAuthenticationEnable(int fd, __u8 Authentication_Enable);
int HCI_ReadEncryptionMode(int fd, __u8 *Encryption_Mode);
int HCI_WriteEncryptionMode(int fd, __u8 Encryption_Mode);
int HCI_ReadClassOfDevice(int fd, __u32 *Class_of_Device);
int HCI_WriteClassOfDevice(int fd, __u32 Class_of_Device);
int HCI_ReadVoiceSetting(int fd, __u16 *Voice_Setting);
int HCI_WriteVoiceSetting(int fd, __u16 Voice_Setting);
int HCI_ReadSCOFlowControl(int fd, __u8 *Flow_Control);
int HCI_WriteSCOFlowControl(int fd, __u8 Flow_Control);
int HCI_ReadLinkSupervisionTimeout(int fd, CHANDLE Connection_Handle, __u16 *Link_Supervision_Timeout);
int HCI_WriteLinkSupervisionTimeout(int fd, CHANDLE Connection_Handle, __u16 Link_Supervision_Timeout);
int HCI_ReadNumberOfSupportedIAC(int fd, __u8 *Num_Supported_IAC);
int HCI_ReadCurrentIACLAP(int fd, __u8 *Num_Current_IAC, __u32 *IAC_LAP);
int HCI_WriteCurrentIACLAP(int fd, __u8 Num_Current_IAC, __u32 *IAC_LAP);


// Informational
int HCI_ReadLocalVersionInformation(int fd, __u8 *HCI_Version, __u16 *HCI_Revision,__u8 *LMP_Version, __u16 *Manufacture_Name, __u16 *LMP_Subversion);
int HCI_ReadLocalSupportedFeatures(int fd, __u64 *LMP_Features);
int HCI_ReadBufferSize(int fd, __u16 *HC_ACL_Data_Packet_Length, __u8 *HC_SCO_Data_Packet_Length, __u16 *Total_Num_ACL_Data_Packets, __u16 *Total_Num_SCO_Data_Packets);
int HCI_ReadCountryCode(int fd, int *Country_Code);
int HCI_ReadBDAddr(int fd, BD_ADDR *bda);


// Ericsson Specific
int HCI_EricssonWritePCMSettings(int fd, __u8 Settings);
int HCI_EricssonSetSCODataPath(int fd, __u8 Path);

#endif
