/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btcore.h,v 1.3 2002/07/25 15:33:20 kds Exp $

   HCI access library

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#ifndef	_BTCORE_H
#define	_BTCORE_H

#include <linux/poll.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#include <affix/hci_types.h>
#include <affix/bluetooth.h>

static inline _syscall3(int,ioctl,int,fd,unsigned int,cmd, void*,arg);
static inline _syscall3(int,poll,struct pollfd*,ufds,unsigned int,nfds,int,timeout);
static inline _syscall2(int,socketcall,int,call,unsigned long*,args);


#ifdef CONFIG_AFFIX_DEBUG
#define dprintf(args...)	DBPRT(DBINIT, args)
#else
#define dprintf(args...)
#endif

static inline int btsys_socket(int domain, int type, int protocol)
{
	unsigned long		a[3];
	int			fd;
	mm_segment_t		old_fs;

	a[0] = domain;
	a[1] = type;
	a[2] = protocol;
	old_fs = get_fs(); set_fs(KERNEL_DS);
	fd = socketcall(SYS_SOCKET, a);
	set_fs(old_fs);

	return fd;
}

static inline int btsys_ioctl(int fd, int cmd, void *arg)
{
	int		err = 0;
	mm_segment_t	old_fs;

	old_fs = get_fs(); set_fs(KERNEL_DS);
	err = ioctl(fd, cmd, arg);
	set_fs(old_fs);

	return err;
}

static inline int btsys_poll(struct pollfd *ufds, unsigned int nfds, int timeout)
{
	int		err;
	mm_segment_t	old_fs;
	
	old_fs = get_fs(); set_fs(KERNEL_DS);
	err = poll(ufds, nfds, timeout);
	set_fs(old_fs);
	
	return err;
}

static inline int btsys_close(int fd)
{
	int		err;
	mm_segment_t	old_fs;
	
	old_fs = get_fs(); set_fs(KERNEL_DS);
	err = close(fd);
	set_fs(old_fs);
	
	return err;
}


static inline int hci_poll(struct pollfd *ufds, unsigned int nfds, int timeout)
{
	return btsys_poll(ufds, nfds, timeout);
}

static inline int hci_close(int fd)
{
	return btsys_close(fd);
}

static inline int hci_exec_cmd(int fd, struct hci_cmd_arg *param)
{
	return btsys_ioctl(fd, BTIOC_EXEC, param);
}

static inline int hci_listen_event(int fd, __u64 mask)
{
	return btsys_ioctl(fd, BTIOC_LISTEN, &mask);
}

static inline int hci_recv_event(int fd, struct hci_event_arg *param)
{
	return btsys_ioctl(fd, BTIOC_RECV_EVENT, param);
}

static inline int hci_recv_msg(int fd, __u8 *param)
{
	return btsys_ioctl(fd, BTIOC_RECV_MSG, param);
}


#endif
