/* 
   Affix - Bluetooth Protocol Stack for Linux

   The contents of this file are subject to the Mozilla Public
   License Version 1.0 (the "License"); you may not use this file
   except in compliance with the License. You may obtain a copy of
   the License at http://www.mozilla.org/MPL/

   Software distributed under the License is distributed on an "AS
   IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   implied. See the License for the specific language governing
   rights and limitations under the License.

   The initial developer of the original code is Mark S. Mathews
   <mark@absoval.com>.  Portions created by Mark S. Mathews
   are Copyright (C) 1998 AbsoluteValue Software, Inc.  All Rights Reserved.

*/

/* 
   $Id: btdebug.h,v 1.1 2002/07/12 07:38:21 kds Exp $

   BTDEBUG - Debug primitives

   This idea has been taken from WLAN driver by Mark S. Mathews.

   Adopted for Affix:
   - some changes
   - parsing

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/

#ifndef _AFFIX_BTDEBUG_H
#define _AFFIX_BTDEBUG_H

#include <affix/types.h>

#include "debug.h"

#ifdef CONFIG_AFFIX_DEBUG
#if 0
static void __get_user_X(void)
{
	return;
}
static __inline__ void __put_user_bad(void)
{
	return;
}
#endif
#endif

extern __u32	affix_dbmask;

#ifdef CONFIG_AFFIX_DEBUG
	#define DBPRT(dbtype, args...) \
	{ \
		if ((affix_dbmask & FILEBIT) && (affix_dbmask & (dbtype))) { \
			if (affix_dbmask & DBFNAME) \
				printk(KERN_DEBUG __FUNCTION__ ": " args); \
			else \
				printk(KERN_DEBUG args); \
		} \
	}

	#define DBPRTF(dbtype, args...) \
	{ \
		if ((affix_dbmask & FILEBIT) && (affix_dbmask & (dbtype))) \
			printk(KERN_DEBUG __FUNCTION__ ": " args); \
	}


	#define DBPRTPKT(dbtype, data, len) \
	{ \
		if( (affix_dbmask & FILEBIT) && (affix_dbmask & DBDUMP) && \
		    (affix_dbmask & (dbtype)) ) \
			printk_hexdump(data, len); \
	}
	
	#define DBPARSEHCI(dbtype, data, len, pkttype, dir) \
	{ \
		if( (affix_dbmask & FILEBIT) && (affix_dbmask & DBPARSE) && \
		    (affix_dbmask & (dbtype)) ) \
			printk_parse_hci(data, len, pkttype, dir); \
	}

	#define DBDUMPCHAR(dbtype, data, len) \
	{ \
		if( (affix_dbmask & FILEBIT) && (affix_dbmask & DBCHARDUMP) && \
		    (affix_dbmask & (dbtype)) ) \
			printk_chardump(data, len); \
	}

	#define DBPARSERFCOMM(dbtype, data, len, dir) \
	{ \
		if( (affix_dbmask & FILEBIT) && (affix_dbmask & DBPARSE) && \
		    (affix_dbmask & (dbtype)) ) \
			printk_parse_rfcomm(data, len, dir); \
	}

#define BTASSERT(expr, func) \
if(!(expr)) { \
        printk( "Assertion failed! %s,%s,%s,line=%d\n",\
        #expr,__FILE__,__FUNCTION__,__LINE__); \
        ##func}


#define DCODE(args...)	{args}

#else /* CONFIG_AFFIX_DEBUG */

#define DBPRT(dbtype, format, args...) 
#define DBPRTF(dbtype, args...) 
#define DBPRTPKT(dbtype, data, len)
#define DBPARSEHCI(dbtype, data, len, pkttype, dir)
#define DBPARSERFCOMM(dbtype, data, len, dir)
#define DBDUMPCHAR(dbtype, data, len)

#define BTASSERT(expr, func)
#define DCODE(args...)

#endif /* CONFIG_AFFIX_DEBUG */

#define DBFENTER	DBPRTF(DBFUNC, "Entered\n")
#define DBFEXIT		DBPRTF(DBFUNC, "Exited\n")

#define BTERROR(args...)   printk(KERN_ERR args)
#define BTERRORF(args...)   printk(KERN_ERR __FUNCTION__ ": " args)
#define BTWARN(args...) printk(KERN_WARNING args)
#define BTWARNF(args...) printk(KERN_WARNING __FUNCTION__ ": " args)
#define BTNOTICE(args...) printk(KERN_INFO args)
#define BTNOTICEF(args...) printk(KERN_INFO __FUNCTION__ ": " args)
#define BTINFO(args...) printk(KERN_INFO args)
#define BTINFOF(args...) printk(KERN_INFO __FUNCTION__ ": " args)


//      For packet parsing
typedef volatile enum
{
        TO_HOSTCTRL = 0,
        FROM_HOSTCTRL
} pkt_dir;

/* some functions to help... */
void printk_hexdump(const __u8 *p, __u32 len);
void printk_chardump(const __u8 *p, int len);

void printk_parse_hci(const __u8 *p, int data_size, int type, pkt_dir dir); 
void printk_parse_rfcomm(const __u8 *p, int len, pkt_dir dir);


#endif
