/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: types.h,v 1.4 2002/08/29 15:37:26 kds Exp $

   TYPEDEF - Common type definitions

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#ifndef TYPEDEF_H
#define TYPEDEF_H

#ifndef __KERNEL__
#include <sys/types.h>
#endif

#include <linux/types.h>

typedef	__u16	PSM;

#define BIT0	0x00000001
#define BIT1	0x00000002
#define BIT2	0x00000004
#define BIT3	0x00000008
#define BIT4	0x00000010
#define BIT5	0x00000020
#define BIT6	0x00000040
#define BIT7	0x00000080
#define BIT8	0x00000100
#define BIT9	0x00000200
#define BIT10	0x00000400
#define BIT11	0x00000800
#define BIT12	0x00001000
#define BIT13	0x00002000
#define BIT14	0x00004000
#define BIT15	0x00008000
#define BIT16	0x00010000
#define BIT17	0x00020000
#define BIT18	0x00040000
#define BIT19	0x00080000
#define BIT20	0x00100000
#define BIT21	0x00200000
#define BIT22	0x00400000
#define BIT23	0x00800000
#define BIT24	0x01000000
#define BIT25	0x02000000
#define BIT26	0x04000000
#define BIT27	0x08000000
#define BIT28	0x10000000
#define BIT29	0x20000000
#define BIT30	0x40000000
#define BIT31	0x80000000

/* Debug levels */
#define DBL_LOG		1
#define DBL_FUNC	2
#define DBL_MEMBER	3
#define DBL_LOCAL	4


#ifdef __GNUC__
	#ifndef __PACK__
		#define __PACK__	__attribute__ ((packed))
	#endif

	#ifndef __INLINE__
		#define __INLINE__	inline
	#endif
#else
	#ifndef __PACK__
		#define __PACK__
	#endif

	#ifndef __INLINE__
		#define __INLINE__ 
	#endif
#endif



#if defined(__i386__)

#define __get_u8(ptr)	(*(__u8*)(ptr))
#define __get_u16(ptr)	(*(__u16*)(ptr))
#define __get_u32(ptr)	(*(__u32*)(ptr))
#define __get_u64(ptr)	(*(__u64*)(ptr))

#define __put_u8(ptr, value)	(*(__u8*)(ptr) = (value))
#define __put_u16(ptr, value)	(*(__u16*)(ptr) = (value))
#define __put_u32(ptr, value)	(*(__u32*)(ptr) = (value))
#define __put_u64(ptr, value)	(*(__u64*)(ptr) = (value))

#else /* not __i386__ */

#if 0

#include <asm/unaligned.h>
#define __get_u8(ptr)	(get_unaligned((__u8*)(ptr)))
#define __get_u16(ptr)	(get_unaligned((__u16*)(ptr)))
#define __get_u32(ptr)	(get_unaligned((__u32*)(ptr)))
#define __get_u64(ptr)	(get_unaligned((__u64*)(ptr)))

#define __put_u8(ptr, value)	(put_unaligned(value, (__u8*)(ptr)))
#define __put_u16(ptr, value)	(put_unaligned(value, (__u16*)(ptr)))
#define __put_u32(ptr, value)	(put_unaligned(value, (__u32*)(ptr)))
#define __put_u64(ptr, value)	(put_unaligned(value, (__u64*)(ptr)))

#else

typedef struct { __u8	value; } __PACK__ __u8_packed;
typedef struct { __u16	value; } __PACK__ __u16_packed;
typedef struct { __u32	value; } __PACK__ __u32_packed;
typedef struct { __u64	value; } __PACK__ __u64_packed;

static inline __u8 __get_u8(void *ptr)
{
	__u8_packed	*p = (__u8_packed*)ptr;
	return p->value;
}
static inline __u16 __get_u16(void *ptr)
{
	__u16_packed	*p = (__u16_packed*)ptr;
	return p->value;
}
static inline __u32 __get_u32(void *ptr)
{
	__u32_packed	*p = (__u32_packed*)ptr;
	return p->value;
}
static inline __u64 __get_u64(void *ptr)
{
	__u64_packed	*p = (__u64_packed*)ptr;
	return p->value;
}

static inline void __put_u8(void *ptr, __u8 value)
{
	__u8_packed	*p = (__u8_packed*)ptr;
	p->value = value;
}
static inline void __put_u16(void *ptr, __u16 value)
{
	__u16_packed	*p = (__u16_packed*)ptr;
	p->value = value;
}
static inline void __put_u32(void *ptr, __u32 value)
{
	__u32_packed	*p = (__u32_packed*)ptr;
	p->value = value;
}
static inline void __put_u64(void *ptr, __u64 value)
{
	__u64_packed	*p = (__u64_packed*)ptr;
	p->value = value;
}

#endif

#endif


/* bluetooth network data types */

#define __btoh16(data)	__le16_to_cpu(data)
#define __btoh32(data)	__le32_to_cpu(data)
#define __btoh64(data)	__le64_to_cpu(data)
#define __htob16(data)	__cpu_to_le16(data)
#define __htob32(data)	__cpu_to_le32(data)
#define __htob64(data)	__cpu_to_le64(data)
#if 0
#define __btoh16p(ptr)	__le16_to_cpup((__u16*)(ptr))
#define __btoh32p(ptr)	__le32_to_cpup((__u32*)(ptr))
#define __btoh64p(ptr)	__le64_to_cpup((__u64*)(ptr))
#define __htob16p(ptr, data)	(*(__u16*)(ptr) = __cpu_to_le16(data))
#define __htob32p(ptr, data)	(*(__u32*)(ptr) = __cpu_to_le32(data))
#define __htob64p(ptr, data)	(*(__u64*)(ptr) = __cpu_to_le64(data))
#endif

/* some affix releated */
#ifndef	__KERNEL__
/* user mode */
//#include <asm/byteorder.h>

#define BT_BYTE_ORDER	__BYTE_ORDER

#else
/* kernel mode */

#ifdef __LITTLE_ENDIAN
#define BT_BYTE_ORDER	__LITTLE_ENDIAN
#endif

#ifdef __BIG_ENDIAN
#define BT_BYTE_ORDER	__BIG_ENDIAN
#endif

#endif

#if BT_BYTE_ORDER == __LITTLE_ENDIAN
#define __btoh24(ptr)	(__btoh32(ptr))
#define __htob24(ptr)	(__htob32(ptr))
#elif BT_BYTE_ORDER == __BIG_ENDIAN
#define __btoh24(ptr)	(__btoh32(ptr)>>8)
#define __htob24(ptr)	(__htob32(ptr)>>8)
#endif

#ifndef btmin
#define btmin(a, b)	(((a)<(b))?a:b)
#endif

#endif /* TYPEDEF_H */
