/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Muller Ulrich <ulrich.muller@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/* 
  $Id: bt_broadcast.c,v 1.2 2002/07/17 11:49:41 kds Exp $
  
  bt_broadcast.h - data structures and functions to send L2CAP packets with Bluetooth Broadcast

   the goal is to send multicast packets efficiently with bluetooth (baseband) broadcast.
   assumptions:
   - the calling unit is a piconet master
   - slaves that are not receivers for the packet but receive it with broadcast silently discard it

*/

/****************************************************************************/


/* kernel includes: */
#include <linux/config.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/vmalloc.h>

/* other includes: */
#define FILEBIT	DBPAN

#include "bt_broadcast.h"
#include "debug.h"

/************************************************************************************************/
/* helpers */

/* test if packet to specified receivers should be sent with broadcast (result != 0) */
int bt_broadcast_test(struct bcast_list *bl, __u32 len) {
	return 0;
}

/* sends a packet to several receivers */
int bt_multicast_send(struct bcast_list *bl, _PSM psm, struct sk_buff *skb_source) {
	int i, err, result;

	if (bl->counter == 0) {
		DBG_F(DBG_MISC, "my job here is done :-)\n");
		dev_kfree_skb(skb_source);
		return 0;
	}

	if(skb_headroom(skb_source) < _L2CAP_hard_header_len ) {
		struct sk_buff *skb_new = skb_realloc_headroom(skb_source, _L2CAP_hard_header_len);
		dev_kfree_skb(skb_source);
		if ((skb_source = skb_new ) == NULL) {
			DBG_F(DBG_MISC, "unable to reallocate buffer\n");
			return -ENOMEM;
		}
	}

	if(bt_broadcast_test(bl, skb_source->len)) { /* use broadcast? */
		DBG_F(DBG_TX_CTRL, "sending with *****BROADCAST*****\n");
		return _L2CA_DataWriteBroadcast(psm, skb_source);
	}

	result = 0;
	i = 0;
	while (i < bl->counter) {
                struct sk_buff *skb;
		if (i + 1 == bl->counter) /* last transmission */
			skb = skb_source;
		else {
			skb = skb_copy(skb_source, GFP_ATOMIC);
			if(!skb) {
				DBG_F(DBG_MISC, "out of memory\n");
				i = bl->counter - 1;
				continue; /* jump to last transmission */
			}
		}

		/* now skb contains a buffer that can be deleted */
		DBG_F(DBG_TX_CTRL, "sending to %s\n", BD_ADDR2str(&bl->con[i]->bda));
		if( (err = _L2CA_DataWrite(bl->con[i], skb)) != 0) {
			DBG_F(DBG_MISC, "error %d\n", err);
			result = err;
		}
		i++;
	}
	return result;
}

/************************************************************************************************/

/*
$Log: bt_broadcast.c,v $
Revision 1.2  2002/07/17 11:49:41  kds
changes

Revision 1.1.1.1  2002/07/11 12:19:04  kds
Initial kernel

Revision 1.7  2002/07/08 08:32:13  kds
changes

Revision 1.6  2002/07/08 08:15:19  kds
changes

Revision 1.5  2002/07/01 11:36:23  kds
l2cap has control info

Revision 1.4  2002/02/07 14:08:36  kds
PAN README

Revision 1.3  2002/02/04 15:19:17  kds
l2cap ...

Revision 1.2  2002/01/09 13:50:25  kds
PAN integrated

Revision 1.1  2002/01/09 12:22:27  kds
PAN added

Revision 1.3  2001/07/31 13:23:42  muelleru
- imported work version

Revision 1.1  2001/07/03 21:21:44  muelleru
- initial version

*/
