/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Muller Ulrich <ulrich.muller@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/* 
 $Id: bt_broadcast.h,v 1.1.1.1 2002/07/11 12:19:04 kds Exp $
 
 bt_broadcast.h - data structures and functions to send L2CAP packets with Bluetooth Broadcast

   the goal is to send multicast packets efficiently with bluetooth (baseband) broadcast.
   assumptions:
   - the calling unit is a piconet master
   - slaves that are not receivers for the packet but receive it with broadcast silently discard it

*/

/************************************************************************************************/
/* includes */

#ifndef _BTBROADCAST_H
#define _BTBROADCAST_H

#include "bt_driver.h"

/************************************************************************************************/
/* types */

struct bcast_list {
	_PSM		psm;		/* destination PSM for broadcast */
	int		counter;	/* number of connection entrys in use */
	_L2CAP_con      *con[7];	/* list of connection handles to target devices */
};

/************************************************************************************************/
/* functions */

/* sends a packet to several receivers, skb will be deleted, headroom of skb will be
   adapted if necessary */
int bt_multicast_send(struct bcast_list *bl, _PSM psm, struct sk_buff *skb);

/************************************************************************************************/

#endif

/*
$Log: bt_broadcast.h,v $
Revision 1.1.1.1  2002/07/11 12:19:04  kds
Initial kernel

Revision 1.3  2002/02/07 14:08:36  kds
PAN README

Revision 1.2  2002/01/09 13:50:25  kds
PAN integrated

Revision 1.1  2002/01/09 12:22:27  kds
PAN added

Revision 1.1  2001/07/03 21:21:44  muelleru
- initial version

*/
